"""A feature extractor combining many sub feature extractors."""

import numpy as np
import pandas as pd

from ...data.field_type import FieldType
from .datetime_feature import DatetimeFeature
from .feature import Feature
from .lag_feature import LagFeature
from .min_feature import MinFeature
from .ordinal_feature import OrdinalFeature
from .skill_feature import SkillFeature
from .total_feature import TotalFeature


class CombinedFeature(Feature):
    """Combined feature extractor class."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        pretrain_features: list[Feature] | None = None,
        posttrain_features: list[Feature] | None = None,
    ) -> None:
        super().__init__()
        if pretrain_features is None:
            pretrain_features = [
                SkillFeature(year_slices=[None, 1, 2, 4, 8]),
                LagFeature(),
                TotalFeature(),
                MinFeature(),
            ]
        if posttrain_features is None:
            posttrain_features = [
                DatetimeFeature(),
                OrdinalFeature(),
            ]
        self._pretrain_features = pretrain_features
        self._posttrain_features = posttrain_features

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self._pretrain_features:
            df = feature.process(df)
        df = df[list(set(df.columns.values) - set(df.attrs[str(FieldType.LOOKAHEAD)]))]
        for feature in self._posttrain_features:
            df = feature.process(df)
        return df.replace([np.inf, -np.inf], np.nan)



"""The prototype class for a game."""

import datetime
from typing import Literal

import pytz
from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType
from .league import League
from .season_type import SeasonType
from .team_model import TeamModel
from .venue_model import VenueModel

GAME_DT_COLUMN = "dt"
SEASON_TYPE_COLUMN = "season_type"
SEASON_YEAR_COLUMN = "year"
VENUE_COLUMN_PREFIX: Literal["venue"] = "venue"
TEAM_COLUMN_PREFIX: Literal["teams"] = "teams"
GAME_ATTENDANCE_COLUMN: Literal["attendance"] = "attendance"
GAME_WEEK_COLUMN: Literal["week"] = "week"


def localize(venue: VenueModel | None, dt: datetime.datetime) -> datetime.datetime:
    """Localize the datetime against the timezone of the venue."""
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        return dt
    tz = None
    venue_model = venue
    if venue_model is not None:
        address = venue_model.address
        if address is not None:
            timezone = address.timezone
            if timezone is not None:
                tz = pytz.timezone(timezone)
    if tz is None:
        tz = pytz.utc
    return tz.localize(dt)


class GameModel(BaseModel):
    """The serialisable representation of a game."""

    dt: datetime.datetime
    week: int | None = Field(..., alias=GAME_WEEK_COLUMN)
    game_number: int | None
    venue: VenueModel | None = Field(..., alias=VENUE_COLUMN_PREFIX)
    teams: list[TeamModel] = Field(..., alias=TEAM_COLUMN_PREFIX)
    end_dt: datetime.datetime | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD}
    )
    attendance: int | None = Field(..., alias=GAME_ATTENDANCE_COLUMN)
    league: League
    year: int | None
    season_type: SeasonType | None



"""The catboost class for a trainer."""

import json
import os
from typing import Any

import optuna
import pandas as pd
import torch
from catboost import CatBoostClassifier  # type: ignore
from catboost import EFeaturesSelectionAlgorithm, EShapCalcType, Pool

# from ...data.columns import GOLDEN_FEATURES_COLUMNS_ATTR
from ..weights import WEIGHTS, CombinedWeight
from .output_column import OUTPUT_COLUMN, output_prob_column
from .trainer import FEATURES_USR_ATTR, HASH_USR_ATTR, Trainer

# from optuna.integration import CatBoostPruningCallback  # type: ignore


_MODEL_FILENAME = "model.cbm"
_USR_ATTR_FILENAME = "usr_attr.json"
# _BORDERS_TSV_FILENAME = "borders.tsv"
_BOOTSTRAP_TYPE_BAYESIAN = "Bayesian"
_BOOTSTRAP_TYPE_BERNOULLI = "Bernoulli"
_OBJECTIVE_LOGLOSS = "Logloss"


def _sanitise_features(x: pd.DataFrame, text_features: list[str]) -> pd.DataFrame:
    # Remove all datetime columns
    x = x[
        [
            column
            for column in x.columns
            if not pd.api.types.is_datetime64_any_dtype(x[column])
        ]
    ]
    x = x[[column for column in x.columns if column not in text_features]]
    return x


class CatboostTrainer(Trainer):
    """The catboost trainer class."""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        folder: str,
        categorical_features: list[str],
        text_features: list[str],
        trial: optuna.trial.Trial | optuna.trial.FrozenTrial | None = None,
    ) -> None:
        # pylint: disable=too-many-locals
        super().__init__(folder, trial=trial)
        self._categorical_features = categorical_features
        self._text_features = text_features
        self._golden_feature_border_count = (
            None
            if trial is None
            else trial.suggest_categorical(
                "golden_feature_border_count", [16, 32, 64, 1028]
            )
        )
        if trial is None:
            self._features_ratio = 0.0
            self._steps = 0
            self._model = CatBoostClassifier(
                task_type=None if not torch.cuda.is_available() else "GPU",
                devices=None if not torch.cuda.is_available() else "0",
            )

        else:
            print("Catboost Trial:")
            print(f"Golden Features Border Count: {self._golden_feature_border_count}")
            self._features_ratio = trial.suggest_float("features_ratio", 0.1, 0.9)
            print(f"Features Ratio: {self._features_ratio}")
            self._steps = trial.suggest_int("steps", 1, 10)
            print(f"Steps: {self._steps}")
            self._weight = CombinedWeight(
                trial.suggest_categorical("weight", list(WEIGHTS.keys()))
            )
            print(f"Weight: {self._weight.weight_name}")
            self._usr_attrs = trial.user_attrs
            bootstrap_type = trial.suggest_categorical(
                "bootstrap_type",
                [_BOOTSTRAP_TYPE_BAYESIAN, _BOOTSTRAP_TYPE_BERNOULLI, "MVS"],
            )
            print(f"Bootstrap Type: {bootstrap_type}")
            bagging_temperature = None
            subsample = None
            if bootstrap_type == _BOOTSTRAP_TYPE_BAYESIAN:
                bagging_temperature = trial.suggest_float(
                    "bagging_temperature", 0.0, 10.0
                )
            elif bootstrap_type == _BOOTSTRAP_TYPE_BERNOULLI:
                subsample = trial.suggest_float("subsample", 0.1, 1.0)
            print(f"Bagging Temperature: {bagging_temperature}")
            print(f"Subsample: {subsample}")
            objective = trial.suggest_categorical(
                "objective", [_OBJECTIVE_LOGLOSS, "CrossEntropy"]
            )
            print(f"Objective: {objective}")
            random_strength = None
            if objective == _OBJECTIVE_LOGLOSS:
                random_strength = trial.suggest_uniform("random_strength", 0.5, 5.0)
            print(f"Random Strength: {random_strength}")
            iterations = trial.user_attrs.get(
                "ITERATIONS", trial.suggest_int("iterations", 100, 10000)
            )
            print(f"Iterations: {iterations}")
            learning_rate = trial.suggest_float("learning_rate", 0.001, 0.3)
            print(f"Learning Rate: {learning_rate}")
            depth = trial.suggest_int("depth", 1, 12)
            print(f"Depth: {depth}")
            l2_leaf_reg = trial.suggest_float("l2_leaf_reg", 3.0, 50.0)
            print(f"L2 Leaf Reg: {l2_leaf_reg}")
            task_type = None if not torch.cuda.is_available() else "GPU"
            print(f"Task Type: {task_type}")
            devices = None if not torch.cuda.is_available() else "0"
            print(f"Devices: {devices}")
            boosting_type = trial.suggest_categorical(
                "boosting_type", ["Ordered", "Plain"]
            )
            print(f"Boosting Type: {boosting_type}")
            self._model = CatBoostClassifier(
                iterations=iterations,
                learning_rate=learning_rate,
                depth=depth,
                l2_leaf_reg=l2_leaf_reg,
                early_stopping_rounds=100,
                # task_type=task_type,
                # devices=devices,
                objective=objective,
                # colsample_bylevel=trial.suggest_float("colsample_bylevel", 0.01, 0.1),
                boosting_type=boosting_type,
                bootstrap_type=bootstrap_type,
                bagging_temperature=bagging_temperature,
                subsample=subsample,
                random_strength=random_strength,
            )

    @property
    def clf(self) -> Any:
        """The underlying classifier"""
        return self._model

    @property
    def salt(self) -> str:
        """The salt to use when hashing the predictions."""
        return "catboost-" + self._usr_attrs[HASH_USR_ATTR]

    def fit(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ):
        """Fit the data."""
        x_train, x_test = x
        y_train, y_test = y
        train_pool = self._create_pool(x_train, y_train, True)  # type: ignore
        eval_pool = None
        if x_test is not None and y_test is not None:
            eval_pool = self._create_pool(x_test, y_test, True)  # type: ignore
        # callbacks = []
        # if self._trial is not None:
        #    callbacks.append(CatBoostPruningCallback(self._trial, "Accuracy"))  # type: ignore
        self._model.fit(
            train_pool,
            eval_set=eval_pool,
            early_stopping_rounds=100,
            # callbacks=callbacks,
        )
        # if callbacks:
        #    callbacks[0].check_pruned()
        feature_importances = self._model.get_feature_importance(prettified=True)
        with pd.option_context("display.max_rows", None, "display.max_columns", None):
            print(feature_importances)

    def save(self):
        """Save the trainer."""
        self._model.save_model(os.path.join(self._folder, _MODEL_FILENAME))
        with open(
            os.path.join(self._folder, _USR_ATTR_FILENAME), "w", encoding="utf8"
        ) as handle:
            json.dump(self._usr_attrs, handle)

    def load(self):
        """Load the trainer."""
        self._model.load_model(os.path.join(self._folder, _MODEL_FILENAME))
        with open(
            os.path.join(self._folder, _USR_ATTR_FILENAME), encoding="utf8"
        ) as handle:
            self._usr_attrs = json.load(handle)

    def predict(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y values."""
        y = super().predict(x)
        if y is not None:
            return y
        train_pool = self._create_pool(x, None, True)
        y = pd.DataFrame(
            index=x.index, data={OUTPUT_COLUMN: self._model.predict(train_pool)}
        )
        self.save_prediction(x, y)
        return y

    def predict_proba(self, x: pd.DataFrame) -> pd.DataFrame | None:
        """Predict the Y probabilities."""
        y = super().predict_proba(x)
        if y is not None:
            return y
        pool = self._create_pool(x, None, True)
        proba = self._model.predict_proba(pool)
        y = pd.DataFrame(
            index=x.index,
            data={output_prob_column(i): proba[:, i] for i in range(proba.shape[1])},
        )
        self.save_prediction_proba(x, y)
        return y

    def select_features(
        self,
        x: tuple[pd.DataFrame, pd.DataFrame | None],
        y: tuple[pd.DataFrame, pd.DataFrame | None],
    ) -> tuple[list[str], int]:
        """Select the features from the training data."""
        x_train, x_test = x
        y_train, y_test = y
        x_train = _sanitise_features(x_train, self._text_features)
        y_train = _sanitise_features(y_train, self._text_features)
        train_pool = self._create_pool(x_train, y_train, False)  # type: ignore
        eval_pool = (
            None
            if x_test is None or y_test is None
            else self._create_pool(x_test, y_test, False)
        )  # type: ignore
        summary = self._model.select_features(
            train_pool,
            num_features_to_select=int(
                self._features_ratio * len(x_train.columns.values)
            ),
            steps=self._steps,
            train_final_model=True,
            features_for_select=x_train.columns.values,
            eval_set=eval_pool,
            algorithm=EFeaturesSelectionAlgorithm.RecursiveByShapValues,
            shap_calc_type=EShapCalcType.Regular,
        )
        return summary["selected_features_names"], self._model.get_best_iteration()

    def _create_pool(
        self,
        x: pd.DataFrame,
        y: pd.DataFrame | None,
        enforce_model_features: bool = False,
    ) -> Pool:
        # pylint: disable=pointless-string-statement
        text_features = list(set(x.columns.values) & set(self._text_features))
        x[text_features] = x[text_features].fillna("").astype(str)
        cat_features = list(set(x.columns.values) & set(self._categorical_features))
        x[cat_features] = x[cat_features].fillna(0).astype(int)
        if enforce_model_features:
            valid_cols = [col for col in self._feature_names if col in x.columns.values]
            x = x[valid_cols]
            text_features = list(set(text_features) & set(valid_cols))
            cat_features = list(set(cat_features) & set(valid_cols))
        x = _sanitise_features(x, self._text_features)
        weight = None
        if y is not None:
            weight = self._weight.process(y)
        pool = Pool(
            x,
            label=y,
            cat_features=cat_features,
            # text_features=text_features,
            weight=weight,
        )
        """
        if self._golden_feature_border_count is not None:
            # pylint: disable=line-too-long
            pool.quantize(
                per_float_feature_quantization=[
                    f"{x.columns.values.tolist().index(feature)}:border_count={self._golden_feature_border_count}"
                    for feature in x.attrs.get(GOLDEN_FEATURES_COLUMNS_ATTR, [])
                    if feature in x.columns.values
                ],
                task_type=None if not torch.cuda.is_available() else "GPU",
            )
            pool.save_quantization_borders(
                os.path.join(self._folder, _BORDERS_TSV_FILENAME)
            )
        else:
            pool.quantize(
                input_borders=os.path.join(self._folder, _BORDERS_TSV_FILENAME),
                task_type=None if not torch.cuda.is_available() else "GPU",
            )
        """
        return pool

    @property
    def _feature_names(self) -> list[str]:
        feature_names = self._model.feature_names_
        if feature_names is None:
            feature_names = self._usr_attrs[FEATURES_USR_ATTR]
        return feature_names



"""The ordinal feature extractor."""

import pandas as pd
from feature_engine.encoding import OrdinalEncoder

from ...data.field_type import FieldType
from .feature import Feature


class OrdinalFeature(Feature):
    """The ordinal feature extractor class."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
    ) -> None:
        super().__init__()
        self._od = OrdinalEncoder(missing_values="ignore", encoding_method="arbitrary")

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        df = self._od.fit_transform(df)
        for categorical_column in df.attrs[str(FieldType.CATEGORICAL)]:
            if categorical_column not in df.columns.values:
                continue
            df[categorical_column] = (
                df[categorical_column].fillna(0).astype(int).astype("category")
            )
        return df



"""A weight based on the combined weight strategies."""

import numpy as np
import pandas as pd

from .class_weight import ClassWeight
from .exponential_weight import ExponentialWeight
from .linear_weight import LinearWeight
from .noop_weight import NoopWeight
from .sigmoid_weight import SigmoidWeight
from .weight import Weight

WEIGHTS = {
    ExponentialWeight.name(): ExponentialWeight,
    LinearWeight.name(): LinearWeight,
    NoopWeight.name(): NoopWeight,
    SigmoidWeight.name(): SigmoidWeight,
}


class CombinedWeight(Weight):
    """Combined weight class."""

    def __init__(self, weight_name: str) -> None:
        self.weight_name = weight_name
        self._class_weight = ClassWeight()
        self._weight = WEIGHTS[weight_name]()

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "combined"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        w1_arr = self._class_weight.process(y)
        w2_arr = self._weight.process(y)
        return w1_arr * w2_arr



"""Tests for the portfolio class."""
import datetime
import shutil
import unittest

import pandas as pd
from sportsball.portfolio.portfolio import Portfolio


class TestPortfolio(unittest.TestCase):

    def setUp(self):
        self._folder = "test_pytest_portfolio"
        self._portfolio = Portfolio([], self._folder)

    def tearDown(self):
        shutil.rmtree(self._folder)

    def test_render(self):
        df = pd.DataFrame(data={
            "afl": [0.0, -0.5, 0.5, 0.2, -0.2],
            "nfl": [0.0, 0.5, -0.5, -0.2, -0.2],
        }, index=[
            datetime.datetime(year=2020, month=1, day=1),
            datetime.datetime(year=2020, month=1, day=2),
            datetime.datetime(year=2020, month=1, day=3),
            datetime.datetime(year=2020, month=1, day=4),
            datetime.datetime(year=2020, month=1, day=5),
        ])
        df.index = df.index.tz_localize("UTC")
        df.index = pd.to_datetime(df.index)
        self._portfolio.render(df)



"""Openmeteo weather model."""

import datetime

import openmeteo_requests  # type: ignore
import pandas as pd
import pytz
import requests
from openmeteo_requests.Client import OpenMeteoRequestsError  # type: ignore

from ...cache import MEMORY
from ..weather_model import WeatherModel


@MEMORY.cache(ignore=["session"])
def create_openmeteo_weather_model(
    session: requests.Session,
    latitude: float,
    longitude: float,
    dt: datetime.datetime,
    tz: str,
) -> WeatherModel:
    """Create a weather model from openmeteo."""
    # pylint: disable=broad-exception-caught
    client = openmeteo_requests.Client(session=session)
    temperature = None
    try:
        responses = client.weather_api(
            "https://historical-forecast-api.open-meteo.com/v1/forecast",
            params={
                "latitude": latitude,
                "longitude": longitude,
                "start_date": str((dt - datetime.timedelta(days=1.0)).date()),
                "end_date": str(dt.date()),
                "hourly": [
                    "temperature_2m",
                    "relative_humidity_2m",
                    "dew_point_2m",
                    "apparent_temperature",
                    "precipitation",
                    "rain",
                    "snowfall",
                    "snow_depth",
                    "weather_code",
                    "pressure_msl",
                    "surface_pressure",
                    "cloud_cover",
                    "cloud_cover_low",
                    "cloud_cover_mid",
                    "cloud_cover_high",
                    "et0_fao_evapotranspiration",
                    "vapour_pressure_deficit",
                    "wind_speed_10m",
                    "wind_speed_100m",
                    "wind_direction_10m",
                    "wind_direction_100m",
                    "wind_gusts_10m",
                    "soil_temperature_0_to_7cm",
                    "soil_temperature_7_to_28cm",
                    "soil_temperature_28_to_100cm",
                    "soil_temperature_100_to_255cm",
                    "soil_moisture_0_to_7cm",
                    "soil_moisture_7_to_28cm",
                    "soil_moisture_28_to_100cm",
                    "soil_moisture_100_to_255cm",
                ],
                "daily": [
                    "weather_code",
                    "temperature_2m_max",
                    "temperature_2m_min",
                    "temperature_2m_mean",
                    "apparent_temperature_max",
                    "apparent_temperature_min",
                    "apparent_temperature_mean",
                    "sunrise",
                    "sunset",
                    "daylight_duration",
                    "sunshine_duration",
                    "precipitation_sum",
                    "rain_sum",
                    "snowfall_sum",
                    "precipitation_hours",
                    "wind_speed_10m_max",
                    "wind_gusts_10m_max",
                    "wind_direction_10m_dominant",
                    "shortwave_radiation_sum",
                    "et0_fao_evapotranspiration",
                ],
                "timezone": tz,
            },
        )
    except (requests.exceptions.RetryError, OpenMeteoRequestsError):
        temperature = None
        return WeatherModel(temperature=temperature)
    except Exception as e:
        e_text = str(e)
        if "Parameter 'start_date' is out of allowed range from" in e_text:
            temperature = None
            return WeatherModel(temperature=temperature)
        raise e
    if not responses:
        temperature = None
        return WeatherModel(temperature=temperature)
    response = responses[0]
    try:
        hourly = response.Hourly()
    except Exception:
        # print(f"Encountered problem unpacking weather: {e}, skipping")
        temperature = None
        return WeatherModel(temperature=temperature)
    if hourly is None:
        raise ValueError("hourly is null.")
    hourly_df = pd.DataFrame(
        index=pd.date_range(
            start=pd.to_datetime(hourly.Time(), unit="s"),
            end=pd.to_datetime(hourly.TimeEnd(), unit="s"),
            freq=pd.Timedelta(seconds=hourly.Interval()),
            inclusive="left",
            tz=tz,
        )[: len(hourly.Variables(0).ValuesAsNumpy())],  # type: ignore
        data={
            "temperature_2m": hourly.Variables(0).ValuesAsNumpy(),  # type: ignore
        },
    )
    dt = dt.replace(tzinfo=None)
    timezone = pytz.timezone(tz)
    dt = timezone.localize(dt)
    hourly_idx = hourly_df.index.get_indexer([dt], method="nearest")[0]
    temperature = hourly_df.iloc[hourly_idx]["temperature_2m"]  # type: ignore
    return WeatherModel(temperature=temperature)



"""NBA ESPN league model."""

import requests

from ...espn.espn_league_model import ESPNLeagueModel
from ...league import League

_SEASON_URL = (
    "http://sports.core.api.espn.com/v2/sports/basketball/leagues/nba/seasons?limit=100"
)


class NBAESPNLeagueModel(ESPNLeagueModel):
    """NBA ESPN implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(_SEASON_URL, League.NBA, session)



"""AFL aussportsbetting bookie model."""

from ....cache import MEMORY
from ...bookie_model import BookieModel


@MEMORY.cache
def create_afl_aussportsbetting_bookie_model() -> BookieModel:
    """Create bookie model from aus sports betting."""
    return BookieModel(identifier="bet365", name="Bet365")



"""AFL aussportsbetting team model."""

from ....cache import MEMORY
from ...team_model import TeamModel
from .afl_aussportsbetting_odds_model import \
    create_afl_aussportsbetting_odds_model


@MEMORY.cache
def create_afl_aussportsbetting_team_model(
    name: str, points: float, odds: float
) -> TeamModel:
    """Create a team model based off aus sports betting."""
    odds_model = create_afl_aussportsbetting_odds_model(odds)
    return TeamModel(
        identifier=name,
        name=name,
        location=None,
        players=[],
        odds=[odds_model],  # pyright: ignore
        points=points,
        ladder_rank=None,
    )



"""The lag feature extractor."""

import pandas as pd

from ...data.league_model import DELIMITER
from .columns import (attendance_column, find_player_count, find_team_count,
                      kick_column, player_identifier_column,
                      team_identifier_column, venue_identifier_column)
from .feature import Feature

LAG_COLUMN_PREFIX = "lag"


def _process_attendance(df: pd.DataFrame) -> pd.DataFrame:
    attendance_col = attendance_column()
    if attendance_col in df.columns.values:
        team_count = find_team_count(df)
        lag_attendance_col = DELIMITER.join([LAG_COLUMN_PREFIX, attendance_col])
        df[lag_attendance_col] = None
        last_attendances: dict[str, int | None] = {}

        def record_lagged_attendence(row: pd.Series) -> pd.Series:
            nonlocal team_count
            nonlocal last_attendances
            venue_idx_col = venue_identifier_column()
            try:
                venue_idx = str(row[venue_idx_col])
            except KeyError:
                return row
            attendance_key_components = [venue_idx]
            for i in range(team_count):
                attendance_key_components.append(row[team_identifier_column(i)])
            attendance_key = "-".join(attendance_key_components)
            row[lag_attendance_col] = last_attendances.get(attendance_key)
            last_attendances[attendance_key] = row[attendance_col]
            return row

        df = df.apply(record_lagged_attendence, axis=1)

    return df


def _process_kicks(df: pd.DataFrame) -> pd.DataFrame:
    team_count = find_team_count(df)
    player_count = find_player_count(df, team_count)
    last_kicks: dict[str, int | None] = {}
    for i in range(team_count):
        for j in range(player_count):
            kick_col = kick_column(i, j)
            lag_kick_col = DELIMITER.join([LAG_COLUMN_PREFIX, kick_col])
            df[lag_kick_col] = None

    def record_team_player_kicks(row: pd.Series) -> pd.Series:
        nonlocal team_count
        nonlocal player_count
        nonlocal last_kicks
        for i in range(team_count):
            for j in range(player_count):
                kick_col = kick_column(i, j)
                lag_kick_col = DELIMITER.join([LAG_COLUMN_PREFIX, kick_col])
                player_idx = None
                try:
                    player_idx = row[player_identifier_column(i, j)]
                except KeyError:
                    continue
                row[lag_kick_col] = last_kicks.get(player_idx)
                try:
                    last_kicks[player_idx] = row[kick_col]
                except KeyError:
                    last_kicks[player_idx] = None
        return row

    return df.apply(record_team_player_kicks, axis=1)


class LagFeature(Feature):
    """The lag feature extractor class."""

    # pylint: disable=too-few-public-methods

    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Process the dataframe and add the necessary features."""
        df = _process_attendance(df)
        df = _process_kicks(df)
        return df.copy()



"""A weight based on linear bias towards the current time."""

import numpy as np
import pandas as pd

from .weight import Weight


class LinearWeight(Weight):
    """Linear weight class."""

    @classmethod
    def name(cls) -> str:
        """The name of the weight class."""
        return "linear"

    def process(self, y: pd.DataFrame) -> np.ndarray:
        return np.linspace(0, 1.0, len(y))



"""Tests for the trainer."""
import unittest

from sportsball.strategy.trainers.trainer import Trainer
import pandas as pd


class TestTrainer(unittest.TestCase):

    def setUp(self):
        self._trainer = Trainer("test_tmp")

    def test_split_train_test(self):
        x = pd.DataFrame(data={"x": [1.0, 0.0, 0.2, 0.3, 0.5, 0.6]})
        y = pd.DataFrame(data={"y": [True, False, True, False, True, True]})
        (x_train, x_test), (y_train, y_test) = self._trainer.split_train_test(x, y)
        self.assertEqual(len(x_train), 4)
        self.assertEqual(len(x_test), 2)
        self.assertEqual(len(y_train), 4)
        self.assertEqual(len(y_test), 2)



"""AFL aussportsbetting league model."""

from io import BytesIO
from typing import Iterator

import requests
import tqdm
from dateutil.parser import parse
from openpyxl import load_workbook

from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from .afl_aussportsbetting_game_model import \
    create_afl_aussportsbetting_game_model


class AFLAusSportsBettingLeagueModel(LeagueModel):
    """AFL AusSportsBetting implementation of the league model."""

    def __init__(self, session: requests.Session) -> None:
        super().__init__(League.AFL, session)

    @property
    def games(self) -> Iterator[GameModel]:
        response = self.session.get(
            "https://www.aussportsbetting.com/historical_data/afl.xlsx"
        )
        response.raise_for_status()
        workbook = load_workbook(filename=BytesIO(response.content))
        ws = workbook.active
        if ws is None:
            raise ValueError("ws is null.")
        for row in tqdm.tqdm(ws.iter_rows(), desc="AusSportsBetting Games"):
            date_cell = str(row[0].value)
            if date_cell in {"Date", "None"}:
                continue
            time_cell = str(row[1].value)
            dt = parse(" ".join([date_cell, time_cell]))
            home_team = str(row[2].value).strip()
            away_team = str(row[3].value).strip()
            venue = str(row[4].value).strip()
            home_points = float(row[5].value)  # type: ignore
            away_points = float(row[6].value)  # type: ignore
            home_odds = float(row[12].value)  # type: ignore
            away_odds = float(row[13].value)  # type: ignore
            yield create_afl_aussportsbetting_game_model(
                dt,
                home_team,
                away_team,
                venue,
                self.session,
                home_points,
                away_points,
                home_odds,
                away_odds,
                self.league,
            )



"""ESPN player model."""

from typing import Any

from ...cache import MEMORY
from ..player_model import PlayerModel


@MEMORY.cache
def create_espn_player_model(player: dict[str, Any]) -> PlayerModel:
    """Create a player model based off ESPN."""
    identifier = str(player["playerId"])
    jersey = player.get("jersey")
    return PlayerModel(identifier=identifier, jersey=jersey, kicks=None)



"""Combined weather model."""

from ..weather_model import WeatherModel


def create_combined_weather_model(
    weather_models: list[WeatherModel],
) -> WeatherModel | None:
    """Create a weather model by combining many weather models."""
    if not weather_models:
        return None
    temperature = None
    for weather_model in weather_models:
        weather_model_temperature = weather_model.temperature
        if weather_model_temperature is not None:
            temperature = weather_model_temperature
    return WeatherModel(temperature=temperature)
